/*
 * This is part of XmlEditor, a graphical editor for XML files
 * Copyright (C) 2001 CERN - Sebastien Ponce
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Contacts :
 *   sebastien.ponce@polytechnique.org
 *   http://lhcb-comp.web.cern.ch/lhcb-comp/
 *   http://www.cern.ch
 */

package ch.cern.lhcb.xmleditor.undo;

import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CannotUndoException;

import org.w3c.dom.Document;

import ch.cern.lhcb.xmleditor.xerces.ElementImpl;

/**
 * this is an implementation of the abstract class AbstractUndoableEdit that
 * defines the undoable attribute eddition in an XmlTree.
 */

public class UndoableAttributeEdition extends AbstractUndoableEdit {

    /**
     * the attribute name
     */
    private String name;

    /**
     * the old attribute value or null if the attribute was not existing
     */
    private String oldValue;

    /**
     * the new attribute value or null if the attribute does not exist anymore
     */
    private String newValue;

    /**
     * the parent of the attribute
     */
    private ElementImpl parent;

    /**
     * a string describing the action for the undo/redo menu items
     */
    private String undoName;

    /**
     * Constructor. Specifies the attribute concerned, its parent and its value
     * before and after the edition
     * 
     * @param document the document this Edit is associated with
     * @param name the name of the attribute
     * @param oldValue the value before the edition
     * @param newValue the value after the edition
     * @param parent the parent of the added node
     * @param undoName the string that will appear in the Edit Menu
     */
    public UndoableAttributeEdition(Document document, String name,
            String oldValue, String newValue, ElementImpl parent,
            String undoName) {
        super(document);
        this.name = name;
        this.oldValue = oldValue;
        this.newValue = newValue;
        this.parent = parent;
        this.undoName = undoName;
    }

    /**
     * Redoes the operation represented by this edit
     * 
     * @throws CannotRedoException if canRedo() returns false
     */
    public void redo() throws CannotRedoException {
        super.redo();
        this.setAttribute(newValue);
    }

    /**
     * Undoes the operation represented by this edit
     * 
     * @throws CannotUndoException if canUndo() returns false
     */
    public void undo() throws CannotUndoException {
        super.undo();
        this.setAttribute(oldValue);
    }

    /**
     * this sets the attribute to the value given. If this value is null, then
     * the attribute is removed
     * 
     * @param value the new value for the attribute (or null to remove it)
     */
    private void setAttribute(String value) {
        if (value == null) {
            // removes the node from its parent
            parent.removeAttribute(name);
        } else {
            // adds the new node to its parent
            parent.setAttribute(name, value, true);
        }
    }

    /**
     * returns an appropriate description of the operation this edit represents.
     * 
     * @return a description of the operation
     */
    public String getPresentationName() {
        return undoName;
    }

}