/*
 * com.simulogic.dataview -- The Java2D Plotting Framework
 * 
 * Copyright (C) 1999 - 2005 Sergey Missan
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this library; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * MThumbSlider: Copyright (C) 1999 Nobuo Tamemasa.
 *  
 */

package com.simulogic.dataview.cursors;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;

import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JToolTip;
import javax.swing.Popup;
import javax.swing.PopupFactory;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.MouseInputAdapter;
import javax.swing.plaf.basic.BasicSliderUI;

public class CursorSliderAdditionalUI {

	CursorSlider slider;

	BasicSliderUI ui;

	Rectangle[] thumbRects;

	int thumbNum;

	private transient boolean isDragging;

	Icon thumbRenderer;

	Rectangle trackRect;

	ChangeHandler changeHandler;

	TrackListener trackListener;

	public CursorSliderAdditionalUI(BasicSliderUI ui) {
		this.ui = ui;
	}

	public void installUI(JComponent c) {
		slider = (CursorSlider) c;
		thumbNum = slider.getThumbNum();
		thumbRects = new Rectangle[thumbNum];
		for (int i = 0; i < thumbNum; i++) {
			thumbRects[i] = new Rectangle();
		}
		isDragging = false;
		trackListener = new CursorSliderAdditionalUI.TrackListener(slider);
		changeHandler = new CursorSliderAdditionalUI.ChangeHandler(slider);
	}

	public void uninstallUI(JComponent c) {
		thumbRects = null;
		trackListener = null;
		changeHandler = null;
	}

	protected void calculateThumbSize() {
		Dimension size = ((CursorSliderAdditional) ui).getThumbSize();
		for (int i = 0; i < thumbNum; i++) {
			if (slider.getOrientation() == SwingConstants.HORIZONTAL) {
				thumbRects[i].setSize(size.width, slider.getSize().height);
			} else {
				thumbRects[i].setSize(slider.getSize().width, size.height);
			}
		}
	}

	protected void calculateThumbLocation() {
		calculateThumbSize();
		for (int i = 0; i < thumbNum; i++) {
			if (slider.getSnapToTicks()) {
				double tickSpacing = slider.getMinorTickSpacing();
				if (tickSpacing == 0) {
					tickSpacing = slider.getMajorTickSpacing();
				}
				if (tickSpacing != 0) {
					double sliderValue = slider.getValueAt(i);
					double snappedValue = sliderValue;
					double min = slider.getMinimumLimit();
					if ((sliderValue - min) % tickSpacing != 0) {
						float temp = (float) (sliderValue - min)
								/ (float) tickSpacing;
						double whichTick = Math.round(temp);
						snappedValue = min + (whichTick * tickSpacing);
						slider.setValueAt(snappedValue, i);
					}
				}
			}
			trackRect = getTrackRect();
			if (slider.getOrientation() == SwingConstants.HORIZONTAL) {
				double valuePosition = ((CursorSliderAdditional) ui)
						.xPositionForValue(slider.getValueAt(i));
				thumbRects[i].x = (int) valuePosition
						- (thumbRects[i].width / 2);
				thumbRects[i].y = slider.getY();
			} else {
				double valuePosition = ((CursorSliderAdditional) ui)
						.yPositionForValue(slider.getValueAt(i));
				thumbRects[i].x = slider.getX();
				thumbRects[i].y = (int) valuePosition
						- (thumbRects[i].height / 2);
			}
		}
	}

	public int getThumbNum() {
		return thumbNum;
	}

	public Rectangle[] getThumbRects() {
		return thumbRects;
	}

	public void setThumbLocationAt(double x, double y, int index) {
		Rectangle rect = thumbRects[index];
		Rectangle unionRect = new Rectangle();
		unionRect.setBounds(rect);

		rect.setLocation((int) x, (int) y);
		SwingUtilities.computeUnion(rect.x, rect.y, rect.width, rect.height,
				unionRect);
		slider.repaint(unionRect.x, unionRect.y, unionRect.width,
				unionRect.height);
	}

	public Rectangle getTrackRect() {
		return ((CursorSliderAdditional) ui).getTrackRect();
	}

	public double xPositionForValue(double value) {
		double min = slider.getMinimumLimit();
		double max = slider.getMaximumLimit();
		double trackLength = trackRect.width;
		double valueRange = (double) max - (double) min;
		double pixelsPerValue = (double) trackLength / valueRange;
		double trackLeft = trackRect.x;
		double trackRight = trackRect.x + (trackRect.width - 1);
		double xPosition;

		if (!drawInverted()) {
			xPosition = trackLeft;
			xPosition += (pixelsPerValue * ((double) value - min));
		} else {
			xPosition = trackRight;
			xPosition -= (pixelsPerValue * ((double) value - min));
		}

		xPosition = Math.max(trackLeft, xPosition);
		xPosition = Math.min(trackRight, xPosition);

		return xPosition;
	}

	public double yPositionForValue(double value) {
		double min = slider.getMinimumLimit();
		double max = slider.getMaximumLimit();
		double trackLength = trackRect.height;
		double valueRange = (double) max - (double) min;
		double pixelsPerValue = (double) trackLength / (double) valueRange;
		double trackTop = trackRect.y;
		double trackBottom = trackRect.y + (trackRect.height - 1);
		double yPosition;

		if (!drawInverted()) {
			yPosition = trackTop;
			yPosition += (pixelsPerValue * ((double) max - value));
		} else {
			yPosition = trackTop;
			yPosition += (pixelsPerValue * ((double) value - min));
		}

		yPosition = Math.max(trackTop, yPosition);
		yPosition = Math.min(trackBottom, yPosition);

		return yPosition;
	}

	/**
	 * Returns a value give a y position. If yPos is past the track at the top
	 * or the bottom it will set the value to the min or max of the slider,
	 * depending if the slider is inverted or not.
	 */
	public double valueForYPosition(double yPos) {
		double value;
		final double minValue = slider.getMinimumLimit();
		final double maxValue = slider.getMaximumLimit();
		final double trackLength = trackRect.height;
		final double trackTop = trackRect.y;
		final double trackBottom = trackRect.y + (trackRect.height - 1);

		if (yPos <= trackTop) {
			value = drawInverted() ? minValue : maxValue;
		} else if (yPos >= trackBottom) {
			value = drawInverted() ? maxValue : minValue;
		} else {
			double distanceFromTrackTop = yPos - trackTop;
			double valueRange = (double) maxValue - (double) minValue;
			double valuePerPixel = valueRange / (double) trackLength;
			double valueFromTrackTop = (distanceFromTrackTop * valuePerPixel);

			value = drawInverted() ? minValue + valueFromTrackTop : maxValue
					- valueFromTrackTop;
		}

		return value;
	}

	/**
	 * Returns a value give an x position. If xPos is past the track at the left
	 * or the right it will set the value to the min or max of the slider,
	 * depending if the slider is inverted or not.
	 */
	public double valueForXPosition(double xPos) {
		double value;
		final double minValue = slider.getMinimumLimit();
		final double maxValue = slider.getMaximumLimit();
		final double trackLength = trackRect.width;
		final double trackLeft = trackRect.x;
		final double trackRight = trackRect.x + (trackRect.width - 1);

		if (xPos <= trackLeft) {
			value = drawInverted() ? maxValue : minValue;
		} else if (xPos >= trackRight) {
			value = drawInverted() ? minValue : maxValue;
		} else {
			double distanceFromTrackLeft = xPos - trackLeft;
			double valueRange = (double) maxValue - (double) minValue;
			double valuePerPixel = valueRange / (double) trackLength;
			double valueFromTrackLeft = (distanceFromTrackLeft * valuePerPixel);

			value = drawInverted() ? maxValue - valueFromTrackLeft : minValue
					+ valueFromTrackLeft;
		}

		return value;
	}

	protected boolean drawInverted() {
		return slider.getInverted();
	}

	public class ChangeHandler implements ChangeListener {

		protected CursorSlider slider;

		ChangeHandler(CursorSlider slider) {
			this.slider = slider;
		}

		public void stateChanged(ChangeEvent e) {
			if (!isDragging) {
				calculateThumbLocation();
				slider.repaint();
			}
		}
	}

	public class TrackListener extends MouseInputAdapter {

		protected transient double offset;

		protected transient double currentMouseX, currentMouseY;

		protected Rectangle adjustingThumbRect = null;

		protected int adjustingThumbIndex;

		protected CursorSlider slider;

		protected Rectangle trackRect;

		Popup tipWindow = null;

		JToolTip tip = null;

		PopupFactory popupFactory = PopupFactory.getSharedInstance();

		private void showTip(MouseEvent e) {
			if (!e.isPopupTrigger()) {
				if (tip == null)
					tip = slider.createToolTip();

				String tipText = slider.getToolTipText(e);
				Point loc = new Point(e.getX(), e.getY());
				SwingUtilities.convertPointToScreen(loc, slider);

				if ((tipText != null) && (tipText.length() > 0)
						&& (loc != null)) {
					tip.setTipText(tipText);
					tipWindow = popupFactory.getPopup(slider.getView(), tip,
							loc.x, loc.y);
					tipWindow.show();
				}
			}
		}

		private void hideTip() {
			if (tipWindow != null)
				tipWindow.hide();
			tipWindow = null;
		}

		TrackListener(CursorSlider slider) {
			this.slider = slider;
		}

		public void mousePressed(MouseEvent e) {
			if (!slider.isEnabled() || e.isShiftDown()) {
				return;
			}
			currentMouseX = e.getX();
			currentMouseY = e.getY();
			slider.requestFocus();

			showTip(e);

			for (int i = 0; i < thumbNum; i++) {
				Rectangle rect = thumbRects[i];
				if (rect.contains(currentMouseX, currentMouseY)) {
					switch (slider.getOrientation()) {
					case SwingConstants.VERTICAL:
						offset = currentMouseY - rect.y;
						break;
					case SwingConstants.HORIZONTAL:
						offset = currentMouseX - rect.x;
						break;
					}
					isDragging = true;
					slider.setValueIsAdjusting(true);
					adjustingThumbRect = rect;
					adjustingThumbIndex = i;
					return;
				}
			}
		}

		public void mouseDragged(MouseEvent e) {
			double value;

			if (!slider.isEnabled() || e.isShiftDown() || e.isControlDown()
					|| !isDragging || !slider.getValueIsAdjusting()
					|| adjustingThumbRect == null) {
				return;
			}
			double thumbMiddle = 0;
			currentMouseX = e.getX();
			currentMouseY = e.getY();
			hideTip();

			Rectangle rect = thumbRects[adjustingThumbIndex];
			trackRect = getTrackRect();
			switch (slider.getOrientation()) {
			case SwingConstants.VERTICAL:
				double halfThumbHeight = rect.height / 2;
				double thumbTop = e.getY() - offset;
				double trackTop = trackRect.y;
				double trackBottom = trackRect.y + (trackRect.height - 1);

				thumbTop = Math.max(thumbTop, trackTop - halfThumbHeight);
				thumbTop = Math.min(thumbTop, trackBottom - halfThumbHeight);

				setThumbLocationAt(rect.x, thumbTop, adjustingThumbIndex);

				thumbMiddle = thumbTop + halfThumbHeight;
				slider.setValueAt(valueForYPosition(thumbMiddle),
						adjustingThumbIndex);
				showTip(e);
				break;

			case SwingConstants.HORIZONTAL:
				double halfThumbWidth = rect.width / 2;
				double thumbLeft = e.getX() - offset;
				double trackLeft = trackRect.x;
				double trackRight = trackRect.x + (trackRect.width - 1);

				thumbLeft = Math.max(thumbLeft, trackLeft - halfThumbWidth);
				thumbLeft = Math.min(thumbLeft, trackRight - halfThumbWidth);

				// setThumbLocationAt(thumbLeft, rect.y, adjustingThumbIndex);

				thumbMiddle = thumbLeft + halfThumbWidth;
				slider.setValueAt(valueForXPosition(thumbMiddle),
						adjustingThumbIndex);

				value = slider.getValueAt(adjustingThumbIndex);
				setThumbLocationAt(xPositionForValue(value) - halfThumbWidth,
						rect.y, adjustingThumbIndex);

				if (slider.isLinked()) {
					int linked = slider.getPartner(adjustingThumbIndex);
					value = slider.getValueAt(linked);

					setThumbLocationAt(xPositionForValue(value)
							- halfThumbWidth, rect.y, linked);
				}

				showTip(e);
				break;
			}
		}

		public void mouseReleased(MouseEvent e) {
			hideTip();

			if (!slider.isEnabled() || e.isShiftDown()) {
				return;
			}
			offset = 0;
			isDragging = false;
			slider.setValueIsAdjusting(false);
		}

		public boolean shouldScroll(int direction) {
			return false;
		}

	}

}